# -*- coding: utf-8 -*-
"""
libs.status
~~~~~~~~~~~~~~~

This module implements WebsiteStatus.

:copyright: (c) 2020 by Liang Hou.
:license: MIT License, see LICENSE for more details.
"""
from http import HTTPStatus
from datetime import datetime
from urllib.parse import urlparse
import msgpack


class WebsiteStatus(dict):
    def __init__(self, status_from, url, status, phrase, dns_time=None, response_time=None, detail='', offset=-1):
        """
        Constructor
        :param status_from: where this test status is from, always transferred to lower cases.
        :param url: Website URL, like https://aiven.io
        :param status: One of "responsive, "unresponsive"
        :param phrase: Status description. This includes all HTTP status code phrases like 'ok', 'forbidden', etc, and
        three additional phrases: 'domain not exist', 'ssl error', and 'connection timeout'.
        :param dns_time: Time in ms taken to resolve DNS.
        :param response_time: Time in ms taken totally to load the Website page.
        :param detail: When phrase is 'Page content not expected', detail shows regular expression for this match.
        :param offset: offset of Kafka message in the corresponding topic
        """
        # Status' UTC timestamp
        timestamp = datetime.utcnow().timestamp()
        # Domain is used as the topic
        self.topic = urlparse(url).netloc.replace('.', '_')
        super().__init__({'from': status_from.lower(), 'url': url, 'timestamp': timestamp,
                          'status': status, 'phrase': phrase, 'dns': dns_time,
                          'response': response_time, 'detail': detail, 'offset': offset})

    @classmethod
    def create_type_schema(cls, db):
        """
        Given a PostgreSQL connection, create enum type for response status and phrases.
        The intention is to save space by using enum type instead of strings directly in DB.
        :param db: A PostgresSQL DB connection instance.
        :return:
        """
        response_status_type_sql = "CREATE TYPE response_status AS ENUM ('responsive', 'unresponsive');"
        # A complete phrases include all HTTP status codes and three additional ones.
        phrases = [getattr(x, 'phrase').lower() for x in HTTPStatus] + ['domain not exist', 'ssl error',
                                                                        'connection timeout',
                                                                        'Page content not expected']
        phrases = tuple(phrases)
        phrase_type_sql = f'CREATE TYPE phrase_status AS ENUM {phrases};'

        with db.cursor() as curs:
            for typename, sql in [('response_status', response_status_type_sql), ('phrase_status', phrase_type_sql)]:
                # Firstly check whether this type exits
                curs.execute(f"select 1 from pg_type where typname = '{typename}'")
                if curs.fetchone():  # Defined already, do nothing
                    continue
                curs.execute(sql)
        db.commit()

    @classmethod
    def create_table_schema(cls, topic, db):
        """
        Given a DB connection, create a corresponding PostgreSQL table for the specific topic.
        :param topic: The specific topic to be created for
        :param db: A PostgresSQL DB connection instance.
        """
        topic = topic.replace('.', '_')
        table_sql = f'''
        CREATE TABLE IF NOT EXISTS web_activity_{topic} (
        id SERIAL PRIMARY KEY,
        created_at TIMESTAMP WITH TIME ZONE DEFAULT CURRENT_TIMESTAMP,
        topic_offset BIGINT DEFAULT -1,
        test_from varchar(32) NOT NULL,
        url varchar(256) NOT NULL,
        event_time BIGINT,
        status response_status,
        phrase phrase_status,
        dns integer DEFAULT -1,
        response integer DEFAULT -1,
        detail varchar(128)
        );
        '''
        index_sql = f'''
        CREATE INDEX IF NOT EXISTS web_activity_url_index_{topic} ON web_activity_{topic} (url);
        '''
        with db.cursor() as curs:
            curs.execute(table_sql)
            curs.execute(index_sql)
        db.commit()

    @classmethod
    def get_last_offset(cls, topic, db):
        """
        Get the last offset received for the specific topic
        :param topic: the topic to get offset for
        :param db: DB connection
        :return: The last offset received from Kafka. -1 if nothing received yet
        """
        topic = topic.replace('.', '_')
        sql = f'''
        SELECT topic_offset from web_activity_{topic} order by id DESC LIMIT 1;
        '''
        with db.cursor() as curs:
            curs.execute(sql)
            r = curs.fetchone()
            if r:
                return r[0]
            else:
                return -1

    def insert_status(self, db):
        """
        Insert this instance to DB as a new row
        :param db: DB connection
        """
        sql = f'''INSERT INTO web_activity_{self.topic} (topic_offset, test_from, url,
        event_time, status, phrase, dns, response, detail) VALUES (
        {self["offset"]}, '{self["from"]}', '{self["url"]}', {self["timestamp"]},
        '{self["status"]}', '{self["phrase"]}', {self["dns"] if self["dns"] else -1},
        {self["response"] if self["response"] else -1}, '{self["detail"] if self["detail"] else ""}'
        );'''
        with db.cursor() as curs:
            curs.execute(sql)
        db.commit()

    def serialize(self):
        """
        Return a serialized bytes
        :return: bytes
        """
        return msgpack.packb(self, use_bin_type=True)

    @classmethod
    def deserialize(cls, raw_bytes):
        """
        Deserialize raw bytes into WebsiteStatus instance
        :param raw_bytes: bytes content
        :return: WebsiteStatus instance
        """
        dictv = msgpack.unpackb(raw_bytes, raw=False)
        reborn = WebsiteStatus(dictv['from'], dictv['url'], dictv['status'], dictv['phrase'], dictv['dns'],
                               dictv['response'], dictv['detail'], dictv['offset'])
        # timestamp has to be set to the original value since reborn comes with a newly created timestamp
        reborn['timestamp'] = dictv['timestamp']
        return reborn



# -*- coding: utf-8 -*-

"""
tracker.tracker
~~~~~~~~~~~~~~~

This module implements WebTracker which sends HTTP GET request to provided URLs and provides their response status.

:copyright: (c) 2020 by Liang Hou.
:license: MIT License, see LICENSE for more details.
"""

import time
import re
import requests
import copy
from urllib.parse import urlparse
from http import HTTPStatus

import dns.resolver

from .status import WebsiteStatus


class WebTracker:
    """
    Class to report Websites Availability.
    """

    def __init__(self, track_from, urls={}, timeout=10):
        """
        Constructor
        :param track_from: Where this Web Availability test is made, like 'sydney'.
        It would be converted to lower cases.
        :param urls: A dictionary with URL and its expected regexp pattern for the page,
        For example, {'https://aiven.io':'<svg\\s+.*', 'https://www.google.com':''} is:
        1. Detect 'https://aiven.io' status and checking whether the response page contains svg tag
        2. Detect 'https://www.goog.com' status
        :param timeout: Seconds to wait for the server to send data before giving up
        """
        self.track_from = track_from.lower()
        # Original dictionary with regexp string
        self.urls = urls
        # This new dictionary with one-off compiled regexp
        self._urls = copy.deepcopy(self.urls)
        for url, reg_exp in urls.items():
            if reg_exp.strip():
                self._urls[url] = re.compile(reg_exp.strip())
        self.timeout = timeout
        # A mapping between HTTP Status codes and their phrases.
        self.status_mapping = dict([(getattr(x, 'value'), getattr(x, 'phrase').lower()) for x in HTTPStatus])

    def status(self):
        """
        A generator to return status for all URLs at the specific period
        :return: A generator yielding WebsiteStatus objects
        """
        for url, re_exp in self._urls.items():
            status = 'unresponsive'
            dns_time = None
            response_time = None
            detail = None

            # We can resolve domain one-off in __init__ but decided not to do so
            # since this is not a time-critical task and it is handy to do in this way.
            domain = urlparse(url).netloc
            try:
                # Measure DNS query time firstly
                stime = time.time()
                # We are measuring DNS response in a rough way through dns.resolver module
                dns.resolver.resolve(domain, 'A')
                etime = time.time()
                # DNS response time in ms
                dns_time = int((etime - stime) * 1000)
                stime = time.time()
                r = requests.get(url, timeout=self.timeout)
                r.close()
                etime = time.time()
                # Total Response time in ms
                response_time = int((etime - stime) * 1000)
                status = 'responsive'
                phrase = self.status_mapping[r.status_code]
                if re_exp:
                    if not re_exp.findall(r.text):
                        phrase = 'Page content not expected'
                        detail = f'{self.urls[url]}'
            except dns.resolver.NXDOMAIN:
                phrase = 'domain not exist'
            except requests.exceptions.SSLError:
                status = 'responsive'
                phrase = 'ssl error'
            except requests.exceptions.ConnectTimeout:
                phrase = 'connection timeout'
            except Exception as e:  # This should never happen, but just in case.
                phrase = str(e).lower()
            yield WebsiteStatus(self.track_from, url, status, phrase, dns_time, response_time, detail)



# -*- coding: utf-8 -*-
"""
unittest for WebsiteStatus

:copyright: (c) 2020 by Liang Hou.
:license: MIT License, see LICENSE for more details.
"""
from contextlib import contextmanager
from http import HTTPStatus
import msgpack
import libs.status
from libs.status import WebsiteStatus


class MockDB:
    """
    A mock DB Connection to verify SQl statement.
    """
    def __init__(self, somethingfetch=False):
        self.sqls = []
        self.somethingfetch = somethingfetch

    class MockCursor:
        def __init__(self, somethingfetch=False):
            self.sqls = []
            self.somethingfetch = somethingfetch

        def execute(self, sql):
            self.sqls.append(sql)

        def fetchone(self):
            if self.somethingfetch:
                return True
            else:
                return False

    @contextmanager
    def cursor(self):
        curs = MockDB.MockCursor(self.somethingfetch)
        yield curs
        self.sqls = curs.sqls

    def commit(self):
        pass


def test_type_schema():
    """
    Verify WebsiteStatus::create_type_schema
    Note: since it is just SQL statement comparison, every whitespace counts. Sigh!
    """
    response_status_type_select_sql = "select 1 from pg_type where typname = 'response_status'"
    response_status_type_sql = "CREATE TYPE response_status AS ENUM ('responsive', 'unresponsive');"
    # A complete phrases include all HTTP status codes and three additional ones.
    phrases = [getattr(x, 'phrase').lower() for x in HTTPStatus] + ['domain not exist', 'ssl error',
                                                                    'connection timeout', 'Page content not expected']
    phrases_type_select_sql = "select 1 from pg_type where typname = 'phrase_status'"
    phrases = tuple(phrases)
    phrase_type_sql = f'CREATE TYPE phrase_status AS ENUM {phrases};'

    db = MockDB()
    WebsiteStatus.create_type_schema(db)
    # Verify 'CREATE TYPE' statement if TYPE phrase_status doesn't exit
    assert db.sqls == [response_status_type_select_sql, response_status_type_sql,
                       phrases_type_select_sql, phrase_type_sql]
    db = MockDB(True)
    WebsiteStatus.create_type_schema(db)
    # Verify no 'CREATE TYPE' statement is executed if TYPE phrase_status exits
    assert db.sqls == [response_status_type_select_sql, phrases_type_select_sql]


def test_table_schema():
    """
    Verify WebsiteStatus::create_table_schema
    Note: since it is just SQL statement comparison, every whitespace counts. Sigh!
    """
    table_sql = '''
        CREATE TABLE IF NOT EXISTS web_activity_aiven_io (
        id SERIAL PRIMARY KEY,
        created_at TIMESTAMP WITH TIME ZONE DEFAULT CURRENT_TIMESTAMP,
        topic_offset BIGINT DEFAULT -1,
        test_from varchar(32) NOT NULL,
        url varchar(256) NOT NULL,
        event_time BIGINT,
        status response_status,
        phrase phrase_status,
        dns integer DEFAULT -1,
        response integer DEFAULT -1,
        detail varchar(128)
        );
        '''
    index_sql = '''
        CREATE INDEX IF NOT EXISTS web_activity_url_index_aiven_io ON web_activity_aiven_io (url);
        '''
    db = MockDB()
    WebsiteStatus.create_table_schema('aiven.io', db)
    assert db.sqls == [table_sql, index_sql]


def test_insert_status():
    """
    Verify WebsiteStatus::insert_status by checking a correct INSERT SQL is submitted to DB.
    Note: since it is just SQL statement comparison, every whitespace counts. Sigh!
    """
    webstatus = WebsiteStatus('Sydney', 'https://aiven.io', 'responsive', 'ok', 10, 300)
    sql = f'''INSERT INTO web_activity_aiven_io (topic_offset, test_from, url,
        event_time, status, phrase, dns, response, detail) VALUES (
        {webstatus["offset"]}, '{webstatus["from"]}', '{webstatus["url"]}', {webstatus["timestamp"]},
        '{webstatus["status"]}', '{webstatus["phrase"]}', {webstatus["dns"] if webstatus["dns"] else -1},
        {webstatus["response"] if webstatus["response"] else -1}, '{webstatus["detail"] if webstatus["detail"] else ""}'
        );'''

    db = MockDB()
    webstatus.insert_status(db)
    assert db.sqls == [sql]


def test_serialize():
    """
    Test WebsiteStatus::serialize
    """
    webstatus = WebsiteStatus('Sydney', 'https://aiven.io', 'responsive', 'ok', 10, 300, '', 100)
    s1 = webstatus.serialize()
    s2 = {'from': 'sydney', 'url': 'https://aiven.io', 'timestamp': webstatus['timestamp'],
          'status': 'responsive', 'phrase': 'ok', 'dns': 10, 'response': 300, 'detail': '', 'offset': 100}
    # webstatus is a dictionary
    assert s1 == msgpack.packb(s2, use_bin_type=True)

    webstatus = WebsiteStatus('Sydney', 'https://aiven.io', 'responsive', 'Page content not expected',
                              10, 300, '<svg\\s+', 100)
    s1 = webstatus.serialize()
    s2 = {'from': 'sydney', 'url': 'https://aiven.io', 'timestamp': webstatus['timestamp'],
          'status': 'responsive', 'phrase': 'Page content not expected', 'dns': 10, 'response': 300,
          'detail': '<svg\\s+', 'offset': 100}
    # webstatus is a dictionary
    assert s1 == msgpack.packb(s2, use_bin_type=True)


def test_deserialize():
    """
    Test WebsiteStatus::deserialize
    """
    # This raw is a msgpack.packb result from WebsiteStatus(15, 'Sydney', 'https://aiven.io', 'responsive',
    #                                                       'forbidden', 10, 300)
    raw = b'\x89\xa4from\xa6sydney\xa3url\xb0https://aiven.io\xa9timestamp\xcbA\xd7\xf9L\x02d.\xa1\xa6status\xaa'\
          b'responsive\xa6phrase\xa9forbidden\xa3dns\n\xa8response\xcd\x01,\xa6detail\xa0\xa6offset\x0f'
    status = WebsiteStatus('Sydney', 'https://aiven.io', 'responsive', 'forbidden', 10, 300, '', 15)
    # Reset timestamp to expected value since timestamp is created lively for every instance.
    # Therefore, it has to be reset to the value for raw
    status['timestamp'] = 1608855561.565346
    assert WebsiteStatus.deserialize(raw) == status
